//
// filemap.cs
//    
//    Implementation of a class to use Win32 MemoryMappedFile
//    objects from .NET applications
//
// COPYRIGHT(C) 2001, Tomas Restrepo (tomasr@mvps.org)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

namespace Tomasr.IO.Test
{
   using System;
   using System.IO;
   using System.Threading;
   using Winterdom.IO;

   //
   // FileMapTest
   //
   /// <summary>
   ///    simple class to test MemoryMappedFiles
   /// </summary>
   public class FileMapTest
   {
      public FileMapTest() 
      {
      }

      public static int Main(string[] args)
      {
         FileMapTest test = new FileMapTest();
      
         if ( args[0] == "/w" )
         {
            // we want to  test with a file write (standalone)
            test.TestMapWrite ( "c:\\filemaptest.bin" );
            
         } else if ( args[0] == "/r" )
         {
            // we want to test a file read  
            test.TestMapRead ( "C:\\filemaptest.bin" );
         } else if ( args[0] == "/c" )
         {
            // try sharing memory as a client 
            test.TestMapShareClient();
         } else if ( args[0] == "/s" )
         {
            // try sharing memory as a server
            test.TestMapShareServer();
         }

         return 0;
      }


      //
      // Testing methods
      //
      private const int kNumViews = 2;
      private const int kFileSize = 128*1024; 
      public void TestMapWrite ( string fileName )
      {
         try
         {
            MemoryMappedFile map = new MemoryMappedFile();
            const int kNumViews = 2;
            const int kFileSize = 128*1024; 
            int viewSize = kFileSize / kNumViews;

            map.Create ( fileName, MapProtection.PageReadWrite, kFileSize );
            

            for ( int i=0; i < kNumViews; i++ )
            {
               Stream stream = map.MapView(MapAccess.FileMapAllAccess, i*viewSize, viewSize );

               BinaryWriter writer = new BinaryWriter(stream);
               for ( int j=0; j < stream.Length; j++ )
                  writer.Write ( (byte)(j%255) );
               
               writer.Flush();
               stream.Close();
            }
            map.Close();
         }
         catch (Exception e)
         {
            Console.WriteLine(e); 
         }
      }

      public void TestMapRead ( string fileName )
      {
      
         try
         {
            MemoryMappedFile map = new MemoryMappedFile();
            int viewSize = kFileSize / kNumViews;

            map.Create ( fileName, MapProtection.PageReadOnly, kFileSize );

            for ( int i=0; i < kNumViews; i++ )
            {
               Stream stream = map.MapView(MapAccess.FileMapRead, i*viewSize, viewSize );

               BinaryReader reader = new BinaryReader(stream);
               const int maxBytes = 8;
               byte[] bytes;
               int offset = i*viewSize;

               for ( int j=0; j < stream.Length; j += maxBytes , offset += maxBytes )
               {
                  bytes = reader.ReadBytes(maxBytes);
                  Console.WriteLine ( 
                     "{0:X8}   {1:X2} {2:X2} {3:X2} {4:X2}  {5:X2} {6:X2} {7:X2} {8:X2}  {9}",
                     offset, bytes[0], bytes[1], bytes[2], bytes[3], bytes[4],
                             bytes[5], bytes[6], bytes[7], BytesToString(bytes) );
               }
               
               stream.Close();
            }
            map.Close();

         }
         catch (Exception e)
         {
            Console.WriteLine(e); 
         }
      }

      private const string kSharedSectionName = "{3f0cef84-e314-4803-b4f0-7b064c051474}";
      private const string kMutexName = "{5bd48c45-9942-4373-a8eb-a972f901da29}";
      private const int kSharedSectionSize = 1024;
      
      public void TestMapShareClient()
      {
         MemoryMappedFile map = new MemoryMappedFile();
         map.Open ( MapAccess.FileMapRead, kSharedSectionName );

         Stream stream = map.MapView(MapAccess.FileMapRead, 0, kSharedSectionSize );
         StreamReader reader = new StreamReader(stream);

         Mutex mutex = new Mutex ( false, kMutexName );

         while ( true )
         {
            mutex.WaitOne();
               string s = reader.ReadLine();
            mutex.ReleaseMutex();
            Console.WriteLine ( s );
            Thread.Sleep(100);
         }

         mutex.Close();
         stream.Close();
         map.Close();
      }
    
      public void TestMapShareServer()
      {
         MemoryMappedFile map = new MemoryMappedFile();
         map.Create ( 
               null, MapProtection.PageReadWrite, 
               kSharedSectionSize, kSharedSectionName 
            );

         Stream stream = map.MapView(MapAccess.FileMapAllAccess, 0, kSharedSectionSize );
         StreamWriter writer = new StreamWriter(stream);

         Mutex mutex = new Mutex ( true, kMutexName );
            mutex.ReleaseMutex();

         while ( true )
         {
            mutex.WaitOne();
               Console.Write ( "> " );
               string s = Console.ReadLine();
               if ( s == "exit" )
                  break;
               writer.WriteLine ( s );
               writer.Flush();
            mutex.ReleaseMutex();
       }

         stream.Close();
         map.Close();
      }

      private string BytesToString(byte[] bytes)
      {
         System.Text.Decoder dec = System.Text.Encoding.ASCII.GetDecoder();
         char[] chars = new char[bytes.Length];
         dec.GetChars(bytes, 0, bytes.Length, chars, 0);
         
         for ( int i = 0; i < chars.Length; i++ )
         {
            if ( !char.IsPrintable(chars[i]) )
               chars[i] = '.';
         }
         return new String(chars);
      }
   
   } // class FileMapTest


} // namespace Tomasr.IO.Test
